"""Common functions used by the rest of the modules."""

from functools import reduce
import numpy as np


def _append_along_0(array1, array2):
    return np.append(array1, array2, axis=0)


def concatenate(a_list):
    """Concatenate the elements of a list."""
    return reduce(_append_along_0, a_list)


def _convolve(array, integral_weights):
    return np.sum(array * integral_weights) / np.sum(integral_weights)


def mfp(cond):
    """The averaged  mean free path (lifetime times |velocity|)"""
    # note that I use the ziman time routine where I however CANCELED
    # in ct.py the forward-scattering correction
    tau = cond._rel_time_ziman()
    return _convolve(tau*cond.vel_norm, cond._integral)


def dos(cond):
    """Density of states at the Fermi level."""
    return np.sum(cond._integral)


def tmfp(cond):
    """The averaged transport mean free path."""
    vmfp_norm = np.linalg.norm(cond.vmfp, axis=1)
    return _convolve(vmfp_norm, cond._integral)


def cdisbalance(cond):
    """Return the valley polarization, see Eq. (33)."""
    n0 = len(cond.geometry.tri[0])
    ns = len(cond.vmfp[n0:, 0])
    cp = 0
    cm = 0
    arange = np.arange(n0, ns)
    mask = cond.kpoints[arange, 0] > 0
    cp = np.sum(cond.vmfp[arange, 0][mask])
    cm = np.sum(cond.vmfp[arange, 0][~mask])
    return (cp - cm) / (ns / 2 * mfp(cond))


def _append_along_0_and_shift(array1, array2):
    if np.ma.is_masked(array1) or np.ma.is_masked(array2):
        append = np.ma.append
    else:
        append = np.append
    return append(array1, array2 + len(array1), axis=0)


def concatenate_and_shift(a_list):
    """Concatenate and shift by the length of the previous element."""
    return reduce(_append_along_0_and_shift, a_list)


def kron(sigma, eta, tau):
    """Triple Kronecker product."""
    return np.kron(sigma, np.kron(eta, tau))


def wrap(array):
    """Wrap the components of a vector into the interval [-pi, pi)."""
    return np.mod(array + np.pi, 2*np.pi) - np.pi


def energy_sorter(wf_array):
    """Sort energies and wave funtions.

    Sort the arrays of energies contiguously, following the highest
    projection of the previous wave function on the new wave functions.
    The result is the set of indices that sort each wave-function in
    `wf_array` with respect to the first wave-function in the array.

    Parameters
    ----------
    wf_array : numpy ndarray or list of numpy arrays
        The wave functions to sort.

    Returns
    -------
    sort : array if ints
        Indices of the sorted array.
    """
    num_vectors = len(wf_array)
    dim = len(wf_array[0])
    sort = np.zeros(shape=(num_vectors, dim), dtype=int)
    sort[0] = np.arange(dim)
    for i_vec in range(1, num_vectors):
        wf0c = np.transpose(np.conjugate(wf_array[i_vec - 1]))[sort[i_vec-1]]
        wf_i = wf_array[i_vec]
        # proyect the wf to the previous wf
        proj = np.abs(np.dot(wf0c, wf_i[:]))
        # pick best wf pairs first, ensuring bijectivity
        found_cols = []
        found_rows = []
        loop = 0
        while ((len(found_cols) < dim or len(found_rows) < dim)
               and loop < dim ** 2):
            loop += 1
            argmax = np.argmax(proj)
            col = argmax // dim
            row = argmax - dim * (argmax // dim)
            proj[col, row] *= 0.
            if col not in found_cols and row not in found_rows:
                sort[i_vec, col] = row
                found_cols.append(col)
                found_rows.append(row)
    return sort


def close_enough(array1, array2, tol=1e-6):
    """Check all elements of arrays."""
    is_close = True
    if np.abs(array1[0] - array2[0]) > tol:
        is_close = False
    elif np.abs(array1[1] - array2[1]) > tol:
        is_close = False
    elif np.abs(array1[2] - array2[2]) > tol:
        is_close = False
    return is_close
