"""
Conductivity Tensor calculation with Boltzmann equation.

Conductivity class
"""

from warnings import warn

import numpy as np
import scipy
import scipy.sparse
import scipy.sparse.linalg as spla
from scipy.special import erf

try:
    from mumps import DMumpsContext
    _mumps_installed = True
except ModuleNotFoundError:
    _mumps_installed = False

import tinyarray as ta

from common import concatenate, concatenate_and_shift
from common import wrap

sigma_x = ta.array([[0, 1], [1, 0]])
sigma_y = ta.array([[0, -1j], [1j, 0]])
sigma_z = ta.array([[1, 0], [0, -1]])
sigma_0 = ta.array([[1, 0], [0, 1]])

_default_amp_params = dict(
    xi=5,
    nI=0.01,
    U=0.1,
    magdis_strength=0.,
    magdis_versor=np.array([0, 0, 1])
)


class Conductivity():
    """Calculate the conductivity tensor of an instance of MarchingCubes class.

    Attributes
    ----------
    amps : sparse matrix in CSR format
        Scattering amplitudes computed with a gaussian potential.
    vmfp : ndarray (N, D)
        Vector mean free paths. 'N' is the number of k-points, and 'D'
        the dimension of the k-space.

    """

    def __init__(self, geometry, rand_spin=False,
                 amp_params=None, tol_gradient=1e-6, amps_cutoff=1e-8):
        """Initialize everything to compute the conductivity.

        Parameters
        ----------
        geometry : instance of MarchingCubes
            Contains the Fermi surface and the relevant
            arrays associated to it.
        amp_params : dict
            Contains parameters necessary to calculate amplitudes
            xi : scalar, correlation length
            nI : scalar, impurities per unit cell
            U : scalar, strength of the (scalar) impurity potential
            magdis_strength : scalar, strength of the spin-dependent (if the
            4-band model is used) or pseudospin-dependent (if 2-band model)
            component of the impurity potential, in units of U
            magdis_versor : array ( shape=(3,) ), direction of the spin-
            dependent potential.
        tol_gradient : scalar, default to 1e-6
            Tolerance for the calculation of the gradient.
        amps_cutoff : float, default to 1e-6
            Cutoff ratio of the amplitudes. Makes the amplitudes and
            rates matrices sparse.
        """
        self.geometry = geometry
        self.rand_spin = rand_spin
        self.amp_params = (_default_amp_params if amp_params is None
                           else amp_params)
        self.model = self.geometry.model
        self.ndim = self.model.ndim
        self.num_contours = len(self.geometry.tri)  #
        self.num_bands = self.model.num_bands
        self.ham_type = self.model.ham_type

        self.kpoints = concatenate(self.geometry.tri)  # in nm^-1
        self.num_points = self.kpoints.shape[0]
        self.wf = concatenate(self.geometry.wf)
        self.vel = concatenate(self.geometry.vel)  # eV*nm = hbar * nm/s
        self.area = concatenate(self.geometry.area)  # in nm^(1-dim)
        self.nei = concatenate_and_shift(self.geometry.nei)
        # calculate amplitudes
        self.amps_cutoff = amps_cutoff
        self._amplitudes()
        self.vel_norm = np.linalg.norm(self.vel, axis=1)  # in eV*nm

        # obtain integral factor over the Fermi surface
        self._integral = self.area / self.vel_norm  # in nm/eV
        # include factor comming from density of states
        self._integral = self._integral / ((2 * np.pi) ** self.ndim)

        # gradient calculation
        self._grad = grad_from_directional_derivatives(
            self.kpoints, self.nei, tol=tol_gradient)

    def __call__(self, mag, all_bands='all', method='sla_solve', dense=True,
                 tol_vmfp=1e-8):
        """Obtain the conductivity tensor.

        Use Ziman (relaxation-time) approximation or the exact solution of
        the Boltzmann equation.

        Parameters
        ----------
        mag : sequence of floats
            Magnetic field in appropriate units.
            This value of magnetic field will be used in the
            triple product 'e / hbar * B.(vel x grad(vector))'
        all_bands : str, one of ['all', 'surface', 'bulk'],  or `None`.
            Whether to sum all contributions to the conductivity.
            If 'surface', sum the conduncitivy contributions form surface
            states only, else, if 'bulk' sum the conductivity contributions
            from the bulk only. If `None`, then return the array of
            conductivity contributions.
        ziman : bool, default to True
            Whether to use Ziman approximation to obtain the
            vector mean-free paths.
        method_solve = str, default to 'sla_solve'
            Method to solve the linear equation of the vmfp equation.
            From all the methods implemented, the fastests (in order)
            are ['sla_solve', 'spla_spsolve', 'mumps'],
            The most reliable and fastest (in order) are
            ['lgmres', 'qmr'].
            Not fast and not reliable are ['bicg', 'bicgstab', 'cgs']
            Other methods that give roughly approximate solutions, and
            fast are ['minres', 'lsqr', 'lsmr'].
            Methods that give good and fast results only when converged
            are ['gcrotmk'].
            At high magnetic fields, the direct solvers are still
            accurate and the fastest (in order): ['sla_solve',
            'spla_spsolve', 'mumps']. However, they raise
            'singular matrix' warnings for dense kpoints
            The iterative solvers become innacurate, except for
            ['qmr']; and the iterative least squares are unreliable.
        dense : bool, default to 'True'
            Use a dense matrix and fix the null space of the rates matrix.
            The dense matrix will be converted to sparse for solvers that
            take a sparse matrix format.
        tol_vmfp : scalar, default to 1e-6
            Relative tolerance for the vector mean-free path convergence.

        """
        self.vmfp = self._get_vmfps(method, mag, tol=tol_vmfp)

        # with the vmfps, calculate the conductivity
        cond = np.zeros((self.num_points, self.ndim, self.ndim))
        for i in range(self.num_points):
            cond[i] = self._integral[i] * np.outer(self.vel[i],
                                                   self.vmfp[i])
        if all_bands == 'all':
            return cond.sum(axis=0)
        elif all_bands == 'surface':
            return cond[np.where(self._weight() > 0.5)[0]].sum(axis=0)
        elif all_bands == 'bulk':
            return cond[np.where(self._weight() <= 0.5)[0]].sum(axis=0)
        return cond

    def _mean_z(self):
        num_layers = self.model.params['layers']
        if self.ham_type == 'weyl2d_fourbands':
            norbs = 4
        else:
            norbs = 2

        assert norbs * num_layers == self.num_bands
        z_operator = np.zeros((self.num_bands, self.num_bands))
        if norbs == 2:
            for i in range(num_layers):
                s = slice(norbs * i, norbs * (i+1))
                z_operator[s, s] = i * sigma_0
        elif norbs == 4:
            for i in range(num_layers):
                s = slice(norbs * i, norbs * (i+1))
                z_operator[s, s] = i * np.kron(sigma_0, sigma_0)

        z_pos = np.zeros((len(self.wf)))
        for k in range(len(self.wf)):
            z_pos[k] = (self.wf[k].conj() @
                        (z_operator @ self.wf[k])).real
        return z_pos

    def _weight(self):
        z_pos = self._mean_z()
        num_layers = self.model.params['layers']

        w = (np.exp(-0.5 * z_pos ** 2)
             + np.exp(-0.5 * (z_pos - num_layers + 1) ** 2))

        w = np.min(np.vstack([w, np.ones_like(w)]), axis=0)  # cap the bottom
        w = np.max(np.vstack([w, np.zeros_like(w)]), axis=0)  # cap the top
        return w

    def _rates(self, method):
        if method == 'ziman':
            tau = self._rel_time_ziman()
            rates = scipy.sparse.diags(1 / tau, format='csr')
        else:
            # compute the rates matrix
            integral_csr = scipy.sparse.diags(self._integral, format='csr')
            # row[i] * integral[i]
            rates = (-self.amps) @ integral_csr
            # average rates should add up to zero, so diagonal is
            # equal to minus the sum of the rows
            sums = self.amps @ self._integral
            rates = rates + scipy.sparse.diags(sums, format='csr')
        return rates

    def _get_vmfps(self, method, mag=0, tol=1e-8):
        rates = self._rates(method)

        # get orbital operator
        omega = self._get_omega(mag * self.model.physical_units['e_over_hbar'])
        # magnetic field after multiplying `e/hbar*B`, when `B` is in Tesla.
        # e*1T/hbar ~ 1.5e-4 * nm**(-2)

        rates_omega = rates + omega

        # rescale the rows to make it more symmetric
        inv_integral_csr = scipy.sparse.diags(1 / self._integral, format='csr')
        # Stabilize the rows with 1 / self._integral.
        # Because of this, the result is multiplied by self._integral
        rates_omega = rates_omega @ inv_integral_csr

        rates_dense = np.array(rates_omega.A)
        null_space_eig = 1
        rates_dense = rates_dense + null_space_eig

        vmfps = np.zeros((self.num_points, self.ndim))

        fermi_contour_length = np.sum(self._integral)

        for i in range(self.ndim):
            b = self.vel[:, i]
            dl = linear_solver(a=rates_dense, b=b,
                               method=method, tol=tol)

            # fix null space
            fix = np.sum(dl) * null_space_eig / (len(dl) * null_space_eig - 1)
            dl = dl - fix

            # recover the original result
            dl = inv_integral_csr @ dl
            # normalize vmfps
            shift = np.dot(dl, self._integral) / fermi_contour_length
            lx = dl - shift
            vmfps[:, i] = lx
        return vmfps

    def _amplitudes(self):
        """Scattering amplitudes."""
        kpoints = self.kpoints
        metric = self.model.lattice_constants
        ham_dim = self.num_bands
        ndim = self.model.ndim

        xi = self.amp_params['xi']  # correlation length
        U = self.amp_params['U']  # potential strength
        nI = self.amp_params['nI']
        # magnetic impurities per unit cell
        magdis_strength = self.amp_params['magdis_strength']
        volume = 1  # volume of the unit cell in nm**ndim

        if self.ham_type == 'weyl2d_fourbands':
            num_bands = 4
            sigmas = [np.kron(sigma_0, sigma_0),
                      np.kron(sigma_0, sigma_x),
                      np.kron(sigma_0, sigma_y),
                      np.kron(sigma_0, sigma_z)
                      ]
        elif self.ham_type == 'weyl2d_twobands':
            num_bands = 2
            sigmas = [sigma_0,
                      sigma_x,
                      sigma_y,
                      sigma_z
                      ]

        layers = int(ham_dim / num_bands)
        z_factor = self._scattering_in_z()
        s = [slice(num_bands * i, num_bands * (i + 1)) for i in range(layers)]

        if magdis_strength == 0:
            V = 1
        else:
            magdis_versor = self.amp_params['magdis_versor']
            mag_dis = magdis_strength * magdis_versor
            V = np.zeros((ham_dim, ham_dim), dtype=complex)

            for i in range(layers):
                V[s[i], s[i]] = np.dot([1, *mag_dis], sigmas)

        # coefficients of the impurity-scattering probabilities calculated
        # in leading-order Born-approximation
        A0 = (2 * np.pi * (np.pi * xi**2) ** ndim * nI * U**2
              / (3 * volume))

        decay = -xi ** 2 / 2
        max_distance_sqr = np.inf
        if self.amps_cutoff > 0:
            max_distance_sqr = np.log(self.amps_cutoff) / decay

        row = []
        col = []
        data = []
        for j in range(self.num_points):
            # first wrap the difference, later multiply by inverse
            # lattice to get the distance in nm again
            distance_sqr = wrap((kpoints[j] - kpoints) * metric)
            distance_sqr = np.sum((distance_sqr / metric) ** 2, axis=1)

            indices = np.nonzero(distance_sqr < max_distance_sqr)[0]
            row.extend([j] * len(indices))
            col.extend(indices)
            distance_sqr = distance_sqr[indices]

            elements = np.exp(decay * distance_sqr)

            # modulation from position in z and z'
            psi = self.wf[indices]
            z_mod = psi.conj() * self.wf[j]
            elements *= (z_mod.conj() * (z_mod @ z_factor)).sum(-1).real

            if self.rand_spin:
                elements *= (1. / np.sqrt(self.wf[j].shape[0]))**2
            else:
                psi = self.wf[indices]
                elements = elements * np.abs(np.dot(
                    psi.conj(), np.dot(V, self.wf[j]))) ** 2

            data.extend(elements)

        amps = scipy.sparse.coo_matrix(
            (data, (row, col)), shape=(self.num_points, self.num_points))
#        amps.setdiag(0)  # remove self-scattering terms

        amps = amps.tocsr() * A0
        self.amps = amps

    def _scattering_in_z(self):
        """Factor that normalizes scattering amplitudes taking into
        account the position of the wavefunctions within the width
        of the slab."""
        w = self.model.params['layers']  # the width is the number of layers
        xi = self.amp_params['xi']  # the impurity range
        if self.ham_type == 'weyl2d_fourbands':
            norbs = 4
        else:
            norbs = 2

        z_array = np.arange(w)
        z_diff = z_array[:, None] - z_array[None, :]
        z_sum = z_array[:, None] + z_array[None, :]
        z_factor = np.exp(-z_diff ** 2 / (4 * xi ** 2))
        z_factor = z_factor / (4 * np.sqrt(np.pi) * xi)
        z_factor = z_factor * (erf((1 + z_sum) / (2 * xi))
                               + erf((2*w - 1 - z_sum) / (2 * xi)))
        # repeat values for every internal degree of freedom
        return np.repeat(np.repeat(z_factor, norbs, axis=0), norbs, axis=1)

    def _rel_time_ziman(self, correction=False):
        """Ziman relaxation time."""
        tau = np.zeros(self.num_points)
        for i in range(self.num_points):
            if correction:
                v_dot_vp = np.dot(self.vel[i, :], np.transpose(self.vel))
                vf_vfp = self.vel_norm[i] * self.vel_norm[:]
                cos_v = 1 - v_dot_vp / vf_vfp
                integral = np.multiply(cos_v, self._integral)
            else:
                integral = self._integral
            denom = self.amps[i].dot(integral)
            if not(denom):
                warn('infinite tau', i)
                tau[i] = np.inf
            else:
                tau[i] = 1 / denom
        return tau

    def _get_omega(self, mag):
        """Return the operator triple product of 'B.(vel x grad)'."""
        orbital = scipy.sparse.csr_matrix((self.num_points, self.num_points))
        cross_prod = np.linalg.norm(-np.cross(self.vel, mag), axis=1)
        orbital = (scipy.sparse.diags(cross_prod, format='csr') @
                   self._grad)

        return orbital


def linear_solver(a, b, method='lgmres', tol=1e-6):
    """Solve 'a x = b'."""
    if method in ['bicg', 'bicgstab', 'cgs', 'gmres', 'lgmres', 'minres',
                  'qmr', 'gcrotmk']:
        solver_type = 'iterative_linear'
    elif method in ['lsqr', 'lsmr']:
        solver_type = 'iterative_least_squares'
    elif method in ['spla_spsolve', 'sla_solve', 'mumps', 'ziman']:
        solver_type = 'direct'
    else:
        raise ValueError('None of the available methods was selected.')

    if solver_type == 'iterative_linear':
        solver = getattr(spla, method)  # method name matches 'spla' method
        x, info = solver(a, b, tol=tol)

        if info < 0:
            raise RuntimeError(f"{method} illegal input or breakdown: {info}")
        if info > 0:
            warn(f'Iterative solver {method}, convergence to tolerance not '
                 f'achieved, number of iterations {info}')

    elif solver_type == 'iterative_least_squares':
        solver = getattr(spla, method)  # method name matches 'spla' method
        x, *status = spla.lsmr(a, b, atol=tol, btol=tol)

        converged, msg = _istop[status[0]]
        if not converged:
            warn(f'Linear solver {method} stops with: {msg}')

    elif solver_type == 'direct':
        # Direct methods for linear equation systems:
        if method == 'spla_spsolve':
            x = spla.spsolve(a, b, use_umfpack=True)
        elif method == 'sla_solve':
            if scipy.sparse.issparse(a):
                x = scipy.linalg.solve(a.A, b)
            else:
                x = scipy.linalg.solve(a, b)
        # Parallel sparse direct solver:
        elif method == 'mumps':
            if scipy.sparse.issparse(a):
                x = _mumps_solver(a, b)
            else:
                x = _mumps_solver(scipy.sparse.coo_matrix(a), b)
        elif method == 'ziman':
            x = a @ b

    return x


def _mumps_solver(a, b):
    if not _mumps_installed:
        raise ValueError("'pymumps' is not installed.")
    ctx = DMumpsContext()
    if ctx.myid == 0:
        ctx.set_centralized_sparse(a)
        x = b.copy()
        ctx.set_rhs(x)  # Modified in place
    ctx.run(job=6)  # Analysis + Factorization + Solve
    ctx.destroy()  # Cleanup
    return x


def grad_from_directional_derivatives(x_array, nei, tol=1e-3):
    """Gradient inferred from the directional derivative.

    `x_array` must be the full dimensional set of points,
    and `nei` the list of neighbours for each point in `x_array`.
    Returns a sparse matrix with the gradient on the direction of the contour.

    The quantity calculated is`grad_U(f)`, where
    `grad_u(f) = u . grad(f)`, with `u` the vector along the
    Fermi surface.

    The derivative is made with the mean of the finite differences to
    the left and to the right of each kpoint.

    This operator is only useful in this form because the velocities
    are always pointing perpendicular to the Fermi surface,
    and the neighbours are in clockwise order. Therefore, the product
    'U x V' is perpendicular to both vectors.

    Otherwise, the direction of the derivative must be included in the
    output.
    """
    num = x_array.shape[0]
    dim = x_array.shape[1]

    mask = np.ma.getmaskarray(nei)  # 'True' for invalid entries
    mask = 1 - mask  # '0' for invalid entries
    num_neighbors_per_k = nei.count(axis=1)  # number of valid entries

    # make finite differences to the left
    dx_0 = np.linalg.norm((x_array - x_array[nei[:, 0]]), axis=1)
    one_over_dx_0 = np.divide(1, dx_0) * mask[:, 0]  # zero out invalid dist.
    # make finite differences to the right
    dx_1 = np.linalg.norm((x_array - x_array[nei[:, 1]]), axis=1)
    one_over_dx_1 = np.divide(1, dx_1) * mask[:, 1]  # zero out invalid dist.

    # when adding the two finite differences we take the mean
    one_over_dx_0 = one_over_dx_0 / num_neighbors_per_k
    one_over_dx_1 = one_over_dx_1 / num_neighbors_per_k

    assert np.all(dx_0 + dx_1 > tol)  # neighbors are sufficiently far away
    assert np.all(one_over_dx_1 + one_over_dx_0 > 0)  # at least one neighbor

    row = np.tile(np.arange(num), dim + 1)
    # central finite differences with different dx to the left and right.
    # (f(x[0]) - f(x[-1])) / (dx_0) + (f(x[1]) - f(x[0])) / (dx_1)
    # -f(x[-1]) / dx_0 + f(x[0]) * (1 / dx_0 - 1 / dx_1) + f(x[1]) / dx_1
    col = np.hstack([nei[:, 0], np.arange(num), nei[:, 1]])
    data = np.hstack([-one_over_dx_0,
                      one_over_dx_0 - one_over_dx_1,
                      one_over_dx_1])
    grad_coo = scipy.sparse.coo_matrix((data, (row, col)), shape=(num, num))
    return grad_coo.tocsr()


# lsqr stopping reason, and if it converged as a bool
_istop = {
    0: (True, 'x=0 is a solution.  If x0 was given, then x=x0 is a solution.'),
    1: (True, 'x is an approximate solution to A*x = B, according to atol and '
        'btol.'),
    2: (True, 'x approximately solves the least-squares problem according to '
        'atol.'),
    3: (False, 'means COND(A) seems to be greater than CONLIM.'),
    4: (True, 'the same as 1 with atol = btol = eps (machine precision)'),
    5: (True, 'is the same as 2 with atol = eps.'),
    6: (False, 'is the same as 3 with CONLIM = 1/eps.'),
    7: (False, 'ITN reached maxiter before the other stopping conditions were '
        'satisfied.')


}
