"""
Conductivity Tensor calculation with Boltzmann equation.

Marching squares and marching cubes algorithms.
"""

from warnings import warn
import numpy as np
import tinyarray as ta


class MarchingSquares():
    def __init__(self, model, bounds, res=10, refine=10,
                 ordered=False,
                 parallel=True, open_contours=True):
        """
        Parameters
        ----------
        model: instance of the `Model` class
            Contains all the relevant functions for obtaining the isolevel
            contours.
        bounds : 2-tuple of 2-tuples
            Bounds in the x-axis and y-axis as
            `((x_min, x_max), (y_min, y_max))`.
        res: int, default to 10
            Number of linear points to subdivide each of the axis intervals.
        refine: int, default to 10
            Number of subdivisions of each square defined by `res`.
        method: 'interpolate' or 'brent'
            Method to find the kpoint that intersects a side of the square.
        parallel : bool, default to 'True'
            Wheather to use parallel calculation of eigenvalues.
        open_contours : bool
            Wheather to allow open contours or raise an error when contours
            do not close on themselves.
        """
        self.parallel = parallel
        self.model = model
        self.bounds = bounds
        if isinstance(res, int):
            self.res = [res, res]
        else:
            # assume res is a tuple or list
            self.res = list(res)
        self.refine = refine
        self.open_contours = open_contours
        self._initial_grid = self._get_initial_grid()

    def __call__(self, f_e=0, tri_only=False):
        """Calcualte the Fermi contours for a Fermi level.

        Sets values for the attributes defined below.

        Parameters
        ----------
        f_e : float, default to '0'
            Fermi level in units of the Hamiltonian eigenvalues.
        tri_only: bool, default to 'False'
            Calculate only the contours of the Fermi surface.

        Attributes
        ----------
        area : list of floats
            Length of segments between a kpoint and the next.
        vel : list of ndarrays
            List of velocities for the kpoints in each band.
            First index refers to the kpoint and the second to the
            component in the 'x, y (z)' dimensions.
        wf : list of ndarrays
            List of wavefunctions for the kpoints in each band.
            First index refers to the kpoint and the second to the
            component in the Hilbert space of the Hamiltonian.
        nei : list of int ndarrays
            Theser arrays are ints and can be used for indexing.
            List of neighbors of each kpoint in each band.
            The numbering is consistent for each band. If the
            bands are later concatenated, the indices must be shifted
            accordingly.
        """
        self.bands = self._iterate_find(f_e)
        self.tri = self.bands

        if not tri_only:
            self._compute_velocities(f_e)
        return self.bands

    @property
    def gridpoints_coarse(self):
        kx1, kx2 = self.bounds[0]
        ky1, ky2 = self.bounds[1]
        n_x, n_y = self.res
        kx_array = np.linspace(kx1, kx2, n_x)
        ky_array = np.linspace(ky1, ky2, n_y)
        return kx_array, ky_array

    @property
    def gridpoints_fine(self):
        kx1, kx2 = self.bounds[0]
        ky1, ky2 = self.bounds[1]
        n_x, n_y = self.res
        kx_array = np.linspace(kx1, kx2, (n_x-1) * self.refine + 1)
        ky_array = np.linspace(ky1, ky2, (n_y-1) * self.refine + 1)
        return kx_array, ky_array

    @property
    def _max_grid_length(self):
        kx1, kx2 = self.bounds[0]
        ky1, ky2 = self.bounds[1]
        n_x, n_y = self.res
        dkx = (kx2 - kx1) / ((n_x-1) * self.refine)
        dky = (ky2 - ky1) / ((n_y-1) * self.refine)
        return np.sqrt(dkx ** 2 + dky ** 2)

    def _iterate_find(self, f_e):
        """Find all the kpoints grouped by contours of the Fermi surface."""
        bands = []
        for iband in range(self.model.num_bands):
            # one band can have multiple contours
            list_of_contours = self._find_contours(iband, f_e)
            for contour in list_of_contours:
                bands.append(np.array(contour))
        new_bands = list(filter(len, bands))
        return new_bands

    def _compute_velocities(self, f_e):
        self.vel = []
        self.wf = []
        self.area = []
        self.nei = []
        for band in self.bands:
            vels, wfs = self._get_vel(band, f_e)
            self.vel.append(vels)
            self.wf.append(wfs)
            neighs = self._calc_neighbours(band)
            self.nei.append(neighs)
            # neighbors first, then area
            self.area.append(self._get_area(neighs, band))

    def _get_vel(self, new_vert, f_e):
        num = len(new_vert)
        num_bands = self.model.num_bands
        ndim = self.model.ndim

        vel = np.zeros(shape=(num, ndim))
        wf = np.zeros(shape=(num, num_bands), dtype=complex)

        for i in range(num):
            vert = new_vert[i]
            this_vel, this_wf = self.model.velocities(vert, e_F=f_e)
            vel[i] = this_vel[0]
            wf[i] = this_wf[:, 0]
        return vel, wf

    def _get_area(self, neighs, new_vert):
        num = len(new_vert)
        areas = np.zeros(shape=(num,))
        mask = np.ma.getmaskarray(neighs)  # 'True' for invalid entries
        mask = 1 - mask
        num_neighbors_per_k = neighs.count(axis=1)  # number of valid entries
        areas = (np.linalg.norm((new_vert - new_vert[neighs[:, 0]]), axis=1)
                 * mask[:, 0])
        areas += (np.linalg.norm((new_vert - new_vert[neighs[:, 1]]), axis=1)
                  * mask[:, 1])
        areas = areas / num_neighbors_per_k

        return areas

    def _calc_neighbours(self, new_vert):
        """In two dimensions the neighbors are just the adjacent points.

        For open contours in 2D, the neighbors have masked entries at the
        endpoints.
        """
        num = len(new_vert)
        nei = np.ma.zeros((num, 2), dtype=int)
        nei[:, 0] = np.roll(np.arange(num), shift=-1)
        nei[:, 1] = np.roll(np.arange(num), shift=1)

        # adjacent kpoints larger than the grid are not neighbors
        max_kdist = self._max_grid_length
        distance_nei_0 = np.linalg.norm(new_vert - new_vert[nei[:, 0]], axis=1)
        nei[distance_nei_0 > max_kdist, 0] = np.ma.masked
        distance_nei_1 = np.linalg.norm(new_vert - new_vert[nei[:, 1]], axis=1)
        nei[distance_nei_1 > max_kdist, 1] = np.ma.masked

        return nei

    def _get_initial_grid(self):
        n_x, n_y = self.res
        kx_array, ky_array = self.gridpoints_coarse
        grid_eigenvalues = np.ndarray((n_x, n_y, self.model.num_bands))
        kx, ky = kx_array[0], ky_array[1]

        if self.parallel:
            from concurrent.futures import ThreadPoolExecutor
            for ix in range(n_x):
                kx = kx_array[ix]

                def eigs_in_ky(ky):
                    nonlocal kx
                    return np.linalg.eigvalsh(self.model.hamiltonian((kx, ky)))
                with ThreadPoolExecutor() as executor:
                    y_map = executor.map(eigs_in_ky, ky_array)
                grid_eigenvalues[ix, :] = np.array([e_wf for e_wf in y_map])
            return grid_eigenvalues

        for ix in range(n_x):
            kx = kx_array[ix]
            for iy in range(n_y):
                ky = ky_array[iy]
                e = np.linalg.eigvalsh(self.model.hamiltonian((kx, ky)))

                grid_eigenvalues[ix, iy] = e[:]
        return grid_eigenvalues

    def _eigs_band(self, iband):
        def f(kx, ky):
            return np.linalg.eigvalsh(
                self.model.hamiltonian((kx, ky)))[iband]
        return f

    def _find_contours(self, iband, f_e):
        func = self._eigs_band(iband)

        gridvals = self._initial_grid[:, :, iband] < f_e
        cells = (
            (gridvals[:-1, :-1] << 0) + (gridvals[1:, :-1] << 1)
            + (gridvals[1:, 1:] << 2) + (gridvals[:-1, 1:] << 3))
        start_points = self._find_one_contour(iband, f_e, cells, func)
        kx_array, ky_array = self.gridpoints_fine
        contours = []
        for start in start_points:
            starting_d_ij = marching_step(cells[start], None, None, None)
            kFs = []
            i, j = ta.array(start) * self.refine
            # d_ij will tell me where I should start in the refined cell
            if starting_d_ij[0] == 1:
                i = i + self.refine
            elif starting_d_ij[1] == 1:
                j = j + self.refine

            cell = MarchingCell((i, j), kx_array, ky_array, func, iso=f_e)

            while cell.cell == 0b0000:
                # move perpendicular to d_ij make sure we intercept the isoline
                if starting_d_ij[0] == 0:
                    cell.move((1, 0))
                elif starting_d_ij[1] == 0:
                    cell.move((0, 1))

            visited_cells = set()
            while True:
                try:
                    kpoint = cell.next()
                except (IndexError, RuntimeError) as _idxerr:
                    if self.open_contours:
                        break
                    else:
                        raise IndexError('Most likely the cell marched into '
                                         'the abyss outside the grid.',
                                         _idxerr)

                this_cell = cell.ij
                if this_cell in visited_cells:
                    break
                kFs.append(kpoint)
                visited_cells.add(this_cell)

            # no empty contours
            if not len(kFs):
                warn(f'Starting cell {start} with direction {starting_d_ij} '
                     'not resolved.')
            else:
                contours.append(np.array(kFs))

        return contours

    def _find_one_contour(self, iband, f_e, cells, func):
        """Get the coarse grid coordinates of one contour, or set of contours

        This is a fast version of the MarchinSquares algorithm, that uses
        the computed values of all cells simultaneously.

        Parameters
        ----------
        iband : int
            Band index of the Fermi surface contour to find.
        f_e : float
            Fermi energy, isolevel of the contours.
        cells : dict
            They keys are tuples of the grid coordinates and the
            values are binaries indicating the type of marching
            squares cell. The grid coordinates correspond to the bottom
            left corner, so the shape is 1 less than that of the
            'self._initial_grid'
        func : callable
            Function used to resolve saddle points, and choose
            a direction for the marching cell to continue.

        """
        # start with a particular cell (needed for later, so that the
        # refinement works easily)
        n_x, n_y = self.res
        kx_array, ky_array = self.gridpoints_coarse

        # construct binary grid of 0's and 1's regions
        grid_binary = 1 * (self._initial_grid[:, :, iband] < f_e)
        # save the values that change after a x-shift
        roll_x = np.logical_xor(grid_binary,
                                np.roll(grid_binary, shift=-1, axis=0))
        # save the values that change after a y-shift
        roll_y = np.logical_xor(grid_binary,
                                np.roll(grid_binary, shift=-1, axis=1))
        # find the union of x-shifts and y-shifts
        roll_xy = np.logical_or(roll_x, roll_y)
        # the indices of the region contours are the `xys`
        xys = {tuple(xy) for xy in np.argwhere(roll_xy)
               if np.all(xy < (n_x - 1, n_y - 1))}  # filter last indices

        start_points = []  # starting points of the contours

        d_ij = None  # no direction yet, but needed for marching step reference

        # we go through all the nontrivial grid points, but not visit the same
        # loop twice.

        while xys:
            initial_point = min(xys)  # lexicographical minimum of tuples
            start_points.append(initial_point)
            xys.discard(initial_point)
            i, j = initial_point
            ij = tuple([i, j])

            middle_k = ((kx_array[i] + kx_array[i+1]) / 2,
                        (ky_array[i] + ky_array[i+1]) / 2)

            while True:
                # make sure we went through all the indices in the contour

                d_ij = marching_step(cells[ij], func, middle_k, d_ij)

                i, j = ta.array(ij) + d_ij
                ij = tuple([i, j])

                xys.discard(ij)  # remove ij if exists

                # check if the next cell exists in the grid
                try:
                    _ = cells[ij]
                except IndexError:
                    if self.open_contours:
                        break
                        raise ValueError(f'The contour {iband} exists the '
                                         'grid values.')

                if ij == initial_point:
                    # The contour closes on itself
                    break
        return start_points


def _cell_values(ij, kxs, kys, f):
    """Takes a tuple of integers 'ij', and evaluates the function 'f'
    at points in '(kxs[i],kys[j])', and suroundings.

    Returns binary representation of the boolean 'values < iso'.
    """
    i, j = ij
    values = (
        f(kxs[i], kys[j]),
        f(kxs[i+1], kys[j]),
        f(kxs[i+1], kys[j+1]),
        f(kxs[i], kys[j+1]),
    )
    return values


def _binary_values(values, iso=0):
    """Returns binary representation of the boolean 'values < iso'.
    """
    cell = (
        ((values[0] < iso) << 0)
        + ((values[1] < iso) << 1)
        + ((values[2] < iso) << 2)
        + ((values[3] < iso) << 3)
    )
    return cell


def marching_cell_values(ij, d_ij, values, kxs, kys, f, iso=0):
    """
    Return the interpolated values where the contour crosses
    the new boundary between squares.

    The order of the 'values' is

    3---2
    |   |
    0---1

    Parameters
    ----------
    ij : pair of ints
        Zeroth position of the cell on the grid.
    d_ij : pair of ints
        Direction where the marching cell is moving.
    values :  quadruple of floats
        Values of the function on the corners around the point 'ij'
        on the grid.
    kxs, kys : ndarray of floats
        Arrays containing the values of the grid in the xy plane
    f : callable
        Function to be evaluated on the grid points near 'ij'.
    iso : float, default to '0'
        The level of the isosurface, that is, contour in 2d.
    """
    i, j = ij
    if d_ij[0] == 1:
        # o-----o-----x
        # | old | new |
        # o-----o-----x
        new_values = (
            values[1],
            f(kxs[i+2], kys[j]),
            f(kxs[i+2], kys[j+1]),
            values[2],
        )
        kx = kxs[i+1]
        weight = (iso - new_values[0]) / (new_values[3] - new_values[0])
        ky = kys[j] + weight * (kys[j+1] - kys[j])
    elif d_ij[0] == -1:
        # x-----o-----o
        # | new | old |
        # x-----o-----o
        new_values = (
            f(kxs[i-1], kys[j]),
            values[0],
            values[3],
            f(kxs[i-1], kys[j+1]),
        )
        kx = kxs[i]
        weight = (iso - new_values[1]) / (new_values[2] - new_values[1])
        ky = kys[j] + weight * (kys[j+1] - kys[j])
    elif d_ij[1] == 1:
        # x-----x
        # | new |
        # o-----o
        # | old |
        # o-----o
        new_values = (
            values[3],
            values[2],
            f(kxs[i+1], kys[j+2]),
            f(kxs[i], kys[j+2]),
        )
        weight = (iso - new_values[0]) / (new_values[1] - new_values[0])
        kx = kxs[i] + weight * (kxs[i+1] - kxs[i])
        ky = kys[j+1]
    elif d_ij[1] == -1:
        # o-----o
        # | old |
        # o-----o
        # | new |
        # x-----x
        new_values = (
            f(kxs[i], kys[j-1]),
            f(kxs[i+1], kys[j-1]),
            values[1],
            values[0],
        )
        weight = (iso - new_values[3]) / (new_values[2] - new_values[3])
        kx = kxs[i] + weight * (kxs[i+1] - kxs[i])
        ky = kys[j]
    else:
        raise ValueError(f"The displacement {d_ij} is trivial.")

    return kx, ky, new_values


class MarchingCell():
    def __init__(self, ij, kxs, kys, eigs, iso=0):
        """
        Parameters
        ----------
        ij : tuple of ints
            Starting index of the contour.
        kxs: numpy array
            Arange of kx values.
        kys: numpy array
            Arange of ky values.
        eigs : callable
            Returns the eigenvalue and eigenfunction of a specific band.
        """
        self.iso = iso
        self.ij = ta.array(ij)
        self.kxs = kxs
        self.kys = kys
        self.eigs = eigs
        self.func = lambda kx, ky: eigs(kx, ky)
        self.d_ij = None

        self.values = _cell_values(self.ij, kxs, kys, self.func)
        self.cell = _binary_values(self.values, iso=self.iso)

    def next(self):
        i, j = self.ij
        middle_point = ((self.kxs[i] + self.kxs[i+1]) / 2,
                        (self.kys[j] + self.kys[j+1]) / 2)

        d_ij = marching_step(self.cell, self.func, middle_point, self.d_ij)

        kx, ky = self._update_cell(d_ij)

        self.d_ij = d_ij
        self.ij = self.ij + d_ij
        return kx, ky

    def move(self, d_ij):
        self._update_cell(d_ij)
        self.d_ij = None
        self.ij = self.ij + ta.array(d_ij)

    def _update_cell(self, d_ij):
        kx, ky, new_values = marching_cell_values(
            self.ij, d_ij, self.values, self.kxs, self.kys, self.func,
            iso=self.iso)

        self.values = new_values
        self.cell = _binary_values(self.values, iso=self.iso)
        return kx, ky


def marching_step(cell, func, middle, d_ij):
    """Return the direction to the next cell


    The parameters `func`, `middle` and `d_ij` are only accessed
    when they are necessary to resolve saddle-point ambiguities
    (i.e. ``cell == 0b0101`` or ``cell == 0b1010``).
    """
    try:
        return MARCHING_STEPS[cell]
    except KeyError:
        msg = "Inconsistent direction and cell values."
        new_d_ij = None
        if cell == 0b0101:
            if func(*middle) < 0:
                if d_ij == (0, 1):
                    new_d_ij = ta.array([1, 0])
                elif d_ij == (0, -1):
                    new_d_ij = ta.array([-1, 0])
                raise RuntimeError(msg)
            else:
                if d_ij == (0, 1):
                    new_d_ij = ta.array([-1, 0])
                elif d_ij == (0, -1):
                    new_d_ij = ta.array([1, 0])
                raise RuntimeError(msg)
        elif cell == 0b1010:
            if func(*middle) < 0:
                if d_ij == (1, 0):
                    new_d_ij = ta.array([0, -1])
                elif d_ij == (-1, 0):
                    new_d_ij = ta.array([0, 1])
                raise RuntimeError(msg)
            else:
                if d_ij == (1, 0):
                    new_d_ij = ta.array([0, 1])
                elif d_ij == (-1, 0):
                    new_d_ij = ta.array([0, -1])
                raise RuntimeError(msg)
        else:
            raise RuntimeError("cell " + bin(cell) + " shouldn't happen ...")

        return new_d_ij


# tables definition
""" MARCHING_STEPS : constant dict
    Assigns a direction to the square corner values.
"""
MARCHING_STEPS = {
    0b0000: ta.array([1, 0]),
    0b0001: ta.array([-1, 0]),
    0b0010: ta.array([0, -1]),
    0b0100: ta.array([1, 0]),
    0b1000: ta.array([0, 1]),
    0b0011: ta.array([-1, 0]),
    0b0110: ta.array([0, -1]),
    0b1100: ta.array([1, 0]),
    0b1001: ta.array([0, 1]),
    0b0111: ta.array([-1, 0]),
    0b1110: ta.array([0, -1]),
    0b1101: ta.array([1, 0]),
    0b1011: ta.array([0, 1])
}
