"""
Conductivity Tensor calculation with Boltzmann equation.

Model class definition and example Hamiltonians
"""

from types import SimpleNamespace
from functools import partial

import numpy as np
import scipy.sparse.linalg

import tinyarray as ta


class _BaseHamiltonian():
    """Class that contains the methods needed for every Hamiltonian system.

    Parameters:
    -----------
    hamiltonian : callable
        Function that should take a 3-tuple as a sole argument,
        and that returns a Hamiltonian.
    vel_method : str ['sparse', 'hf'], default to 'sparse'
        Method for finding the velocities and wavefuctions. If 'sparse',
        then the velocities are calculated numerically using a forward
        partial derivate with spacing 'dk'. This is the most efficient method,
        and represents no issues in Fermi surfaces with no degeneracies.
        If 'hf', then the velocities obtained with Hellmann-Feynmann theorem,
        by computing the eigenvectors and the partial derivate with spacing
        'dk' of the Hamiltonian.
    dk : scalar, default to 1e-5
        Spacing in k-space for calculating partial derivates of the
        Hamiltonian.
    lattice_constants : tuple or array, default to np.array([1,1,1])
        Lattice dimensions in the `x`, `y` and `z` directions.
    ndim : int, default: 3
        Dimension of the reciprocal space.
    """

    def __init__(self, hamiltonian=None, dk=1e-5, vel_method='sparse',
                 lattice_constants=None, ndim=3):
        self.ndim = ndim
        self.dk = dk
        assert callable(hamiltonian)
        self.lattice_constants = lattice_constants or ta.array([1] * self.ndim)
        self.hamiltonian = hamiltonian
        self.num_bands = hamiltonian(ta.array([0.] * self.ndim)).shape[0]
        self.eigenvals = lambda k: np.linalg.eigh(self.hamiltonian(k))
        self.vel_method = vel_method

    def _eigs_near_eF(self, k, e_F=0, nbands=1, return_eigenvectors=True):
        ham_k = self.hamiltonian(k)
        if self.vel_method == 'sparse':
            return scipy.sparse.linalg.eigs(
                ham_k, sigma=e_F, k=nbands,
                return_eigenvectors=return_eigenvectors)
        else:
            if return_eigenvectors:
                eigs, wfs = np.linalg.eigh(ham_k)
                sort = np.argsort(np.abs(eigs - e_F))[:nbands]
                return eigs[sort], wfs[:, sort]
            else:
                eigs = np.linalg.eigvalsh(ham_k)
                sort = np.argsort(np.abs(eigs - e_F))[:nbands]
                return eigs[sort]

    def velocities(self, k, e_F=0, num_bands=None):
        """Calculate the velocity at point 'k'.

        If 'self.method == "sparse"', the velocities are computed around
        'e_F'. In this case, if 'num_bands' is not provided, the default
        number of bands is '1'.
        If 'self.method == "hf"' and 'num_bands' is not provided, the default
        number of bands is all the bands 'self.num_bands'.

        Therefore, if 'num_bands' is not provided, the shape of the arrays
        in the output have different shape depending on the value of
        'self.method'.

        When many bands are computed, the velocities are calculated using
        the eigenvalues in ascending order.
        """
        dk = self.dk
        nbands = 1 if num_bands is None else num_bands
        e, wf = self._eigs_near_eF(k, e_F=e_F, nbands=nbands,
                                   return_eigenvectors=True)
        e = np.array(sorted(np.real(e)))
        vel = np.zeros(shape=(nbands, self.ndim))

        for idim in range(self.ndim):
            # first derivates
            dk_array = np.array([0.] * self.ndim)
            dk_array[idim] = dk
            if self.vel_method == 'hf':
                dh_dk = (self.hamiltonian(k + dk_array)
                         - self.hamiltonian(k))
                for iband in range(nbands):
                    v = np.dot(wf[:, iband].conj().T,
                               np.dot(dh_dk, wf[:, iband])).real
                    vel[iband, idim] = v / dk
            elif self.vel_method == 'sparse':
                e_dk = self._eigs_near_eF(k + dk_array, e_F=e_F, nbands=nbands,
                                          return_eigenvectors=False)
                e_dk = np.array(sorted(np.real(e_dk)))
                vel[:, idim] = (e_dk - e) / dk
            else:
                raise NotImplementedError()
        return vel, wf


class Model(_BaseHamiltonian):
    """Contains all the necessary parameters of a model

    Includes the Hamiltonian of the system and methods associated to it.
    Also the physical  units and magnetic field.

    Parameters:
    -----------
        ham : callable, optional
            Hamiltonian with signature `(k, mag=0)`.
        params : list or tuple
            Parameters to pass to the Hamiltonian
        ham_type : str
            Hamiltonian of the system, one of 'free', 'sio', 'weyl',
            'weyl2' or 'fighter'.
        dk : scalar, default to 1e-5
            Spacing in k-space for calculating partial derivates of the
            Hamiltonian.
        mag_versor : tuple or array, default to np.array([0,0,1])
            Direction of the magnetic field.
        mag_intensity : scalar, default to 0.
            Intensity of the magnetic field.
        vel_method : str, optional, default to 'sparse'
            Method for calculating velocities. If 'sparse', the scipy
            sparse eigenvalue solver is used to compute forward partial
            derivatives with spacing 'dk'.
            If 'hf' the velocities obtained with Hellmann-Feynmann theorem.
    """

    def __init__(self, ham=None, params=[], ham_type='free', dk=1e-5,
                 mag_versor=np.array([0, 0, 1]), vel_method='sparse', ndim=3):
        self.ndim = ndim
        self.params = params
        lattice_constants = ta.array([1] * self.ndim)
        if ham_type == 'free':
            ham = ham_free
        elif ham_type == 'sio':
            ham = ham_sio
            lattice_constants = np.array([0.556, 0.559, 0.788])
        elif ham_type == 'weyl':
            ham = ham_weyl
        elif ham_type == 'weyl2':
            ham = ham_weyl2
        elif ham_type == 'weyl2d_twobands':
            ham = ham_weyl_2d_2bands
        elif ham_type == 'weyl2d_fourbands':
            ham = ham_weyl_2d_4bands
        else:
            if ham is None:
                raise ValueError('`ham` not provided')

        # a string
        self.ham_type = ham_type

        self.physical_units = dict(
            mu_b=5.788e-5,  # Bohr magneton in eV*T-1
            g_factor=2,  # Landé g-factor
            mag_eV_over_T=2 * 5.788e-5,  # g_factor * mu_b
            hbar=6.582119514e-16,  # in eV * s
            e_over_hbar=1e-18/6.582119514e-16,  # e*T/hbar in nm^-2
            # this is equal to eV*s/(hbar * m^2), because e*T = eV*s/m^2
        )

        self.suffix = (self.ham_type + '.p')

        # magnetic field in eV
        hamiltonian = partial(ham, params=params)

        _BaseHamiltonian.__init__(self, hamiltonian=hamiltonian,
                                  lattice_constants=lattice_constants, dk=dk,
                                  vel_method=vel_method, ndim=ndim)


sigma_x = tau_x = nu_x = ta.array([[0, 1], [1, 0]])
sigma_y = tau_y = nu_y = ta.array([[0, -1j], [1j, 0]])
sigma_z = tau_z = nu_z = ta.array([[1, 0], [0, -1]])
sigma_0 = tau_0 = nu_0 = ta.array([[1, 0], [0, 1]])


def kron(sigma, nu, tau):
    """Triple Kronecker product"""
    return np.kron(sigma, np.kron(nu, tau))


def ham_sio(k, mag=0):
    """Hamiltonian of SIO at kx, ky, kz """
    kx, ky, kz = k

    p = SimpleNamespace(
        l0=0.,  # SOC, diagonal in the Hamiltonian
        t=-0.6,  # nearest neighbors intra-orbital hoppings
        tp=-0.15,  # hopping between 1d orbitals
        t_xy=-0.3,  # nearest neighbors in-plane hoppings
        t0z=0.13,  # 1d to xy hoppings between adjacent atoms on
        # different planes
        td=-0.3,  # intra-orbital hoppings of NNN from B to G
        tpd=0.03,  # intra-orbital hoppings of NNN from R to Y
        t01p=0.1,  # NN in-plane hoppings between 1d and xy orbitals
        t02p=0.3,  # NN in-plane hoppings between 1d and xy orbitals
        t0d=0.06)

    def E0(kx, ky, kz): return p.l0 + p.t_xy * np.cos(kx) * np.cos(ky)
    def Ep(kx, ky, kz): return (2 * p.t - 1j * p.tp) * \
        (np.cos(kx) + np.cos(ky))

    def Ez(kx, ky, kz): return 2 * p.t * np.cos(kz)
    def Ez0(kx, ky, kz): return (1 - 1j) * p.t0z * np.cos(kz)

    def Ed(kx, ky, kz): return p.td * (np.cos(kx) + np.cos(ky)) * np.cos(kz) \
        + 1j * p.tpd * (np.sin(kx) + np.sin(ky)) * np.sin(kz)

    def Ep0(kx, ky, kz): return (p.t01p * np.cos(kx) + p.t02p * np.cos(ky)) \
        - 1j * (p.t02p * np.cos(kx) + p.t01p * np.cos(ky))
    def Ed0(kx, ky, kz): return (
        p.t0d * (np.sin(ky) + 1j * np.sin(kx)) * np.sin(kz))

    # hamiltonian calculation
    ham = np.zeros((8, 8), dtype=complex)
    ham += kron(sigma_0, nu_0, tau_0) * E0(kx, ky, kz)
    ham += kron(sigma_0, nu_0, tau_x) * np.real(Ep(kx, ky, kz))
    ham += kron(sigma_z, nu_0, tau_y) * np.imag(Ep(kx, ky, kz))
    ham += kron(sigma_0, nu_x, tau_0) * Ez(kx, ky, kz)
    ham += kron(sigma_0, nu_x, tau_x) * np.real(Ed(kx, ky, kz))
    ham += kron(sigma_0, nu_y, tau_y) * np.imag(Ed(kx, ky, kz))
    ham += kron(sigma_y, nu_z, tau_y) * np.real(Ep0(kx, ky, kz))
    ham += kron(sigma_x, nu_z, tau_y) * np.imag(Ep0(kx, ky, kz))
    ham += kron(sigma_y, nu_y, tau_z) * np.real(Ez0(kx, ky, kz))
    ham += kron(sigma_x, nu_y, tau_z) * np.imag(Ez0(kx, ky, kz))
    ham += kron(sigma_y, nu_x, tau_y) * np.real(Ed0(kx, ky, kz))
    ham += kron(sigma_x, nu_x, tau_y) * np.imag(Ed0(kx, ky, kz))
    ham /= 3
    if np.any(mag) > 0:
        ham += mag[0] * kron(sigma_x, tau_0, nu_0)
        ham += mag[1] * kron(sigma_y, tau_0, nu_0)
        ham += mag[2] * kron(sigma_z, tau_0, nu_0)
    return ham


def ham_weyl(k, mag=0):
    """Minimal model of a Weyl semimetal

    Taken from `https://arxiv.org/abs/1604.03096`_.
    """
    kx, ky, kz = k
    t = 1
    tx = t
    k0 = 1.  # location of the Weyl nodes in the kx axis
    m0 = 2
    m = m0 * (2 - np.cos(ky) - np.cos(kz)) + 2 * tx * (np.cos(kx) - np.cos(k0))
    gamma = 0  # type I Weyl semimetal if gamma < 2 * tx

    ham = (gamma * (np.cos(kx) - np.cos(k0)) * sigma_0
           - m * sigma_x
           - 2 * t * np.sin(ky) * sigma_y
           - 2 * t * np.sin(kz) * sigma_z
           )

    if np.any(mag) > 0:
        ham += mag[0] * sigma_x
        ham += mag[1] * sigma_y
        ham += mag[2] * sigma_z
    return ham


def ham_weyl2(k, mag=0):
    """Weyl model Hamiltonian

    Created following the equations in
    `https://arxiv.org/abs/1512.02144`_."""
    kx, ky, kz = k
    coskx = np.cos(kx)
    cosky = np.cos(ky)
    coskz = np.cos(kz)
    sinkx = np.sin(kx)
    sinky = np.sin(ky)

    # hbar2_over2m in eV over nm
    t = 1  # nearest neighbor hopping in the xy plane
    tz = t
    tp = 2 * t  # hopping with SO
    tzp = tz  # hoppint with SO
    m0 = 0.3 * t
    gamma = 0.  # strain
    beta = 1.2 * t  # magnetization

    m = m0 + t * (2 - coskx - cosky) + tz * (1 - coskz)

    ham = (
        tp * sinkx * np.kron(tau_z, sigma_x)
        + tp * sinky * np.kron(tau_z, sigma_y)
        + tzp * np.kron(tau_y, sigma_0)
        + m * np.kron(tau_x, sigma_0)
        + 0.5 * gamma * np.kron(tau_y, sigma_z)
        + 0.5 * beta * np.kron(tau_0, sigma_z)
    )
    if np.any(mag) > 0:
        ham += mag[0] * np.kron(tau_0, sigma_x)
        ham += mag[1] * np.kron(tau_0, sigma_y)
        ham += mag[2] * np.kron(tau_0, sigma_z)
    return ham


def ham_free(k, v_f=1):
    """Free electron hamiltonian with a potential."""
    k2 = v_f * np.linalg.norm(k)**2
    return ta.array([[k2,  0],
                     [0, -k2]])


def ham_weyl_2d_2bands(k, params=None):
    """Two-band hamiltonian of a Weyl semimetal without time-reversal
    symmetry.

    Parameters
    ----------
    params: dict
        Can contain the parameters
        layers: int, default to `1`
            Number of layers in the slab.
        pbc : bool, default to False
            If 'True', use periodic boundary conditions, and the slab will
            not host surface states.
        p: SimpleNamespace, optional
            Contains the Hamiltonian parameters,
                p.mu : chemical potential
                p.mup : chemical potential at the surface
                p.t : (spin-orbit) hopping parameter (isotropic in this model)
                p.tp : Magnetic field in the y direction
                p.k0 : half the k-space separation between Weyl nodes
                p.lam : energy shift between Weyl nodes
    """
    if params is None:
        params = {'layers': 1}
    elif len(params) == 0:
        raise ValueError('`params` must be provided.')
    layers = params['layers']
    pbc = params.get('pbc', False)
    assert layers > 0
    kx, ky = k
    try:
        p = params['p']
    except KeyError:
        p = SimpleNamespace(mu=0.001, t=1, tp=1, k0=1.2, lam=0.2,
                            Bpar=0., Bperp=0.)
    ham_dim = 2 * layers

    ham = np.zeros((ham_dim, ham_dim), dtype=complex)
    s = [slice(2*i, 2*i+2) for i in range(layers)]
    for i in range(layers):
        if i in [0, layers-1]:
            onsite = -p.mup * sigma_0
        else:
            onsite = -p.mu * sigma_0
        ham[s[i], s[i]] = (
            onsite + p.t * np.sin(ky+(i-(layers-1)/2)*p.Bpar) * sigma_y
            + (p.t * (2 - np.cos(ky+(i-(layers-1)/2)*p.Bpar) -
                      np.cos(kx-(i-(layers-1)/2)*p.Bperp))
               + p.t * np.cos(p.k0))
            * sigma_z
        )

    for i in range(layers - 1):
        ham[s[i], s[i+1]] = -0.5j * p.t * sigma_x - 0.5 * p.t * sigma_z
        ham[s[i+1], s[i]] = 0.5j * p.t * sigma_x - 0.5 * p.t * sigma_z

    if pbc:
        ham[s[-1], s[0]] = -0.5j * p.t * sigma_x - 0.5 * p.tp * sigma_z
        ham[s[0], s[-1]] = 0.5j * p.t * sigma_x - 0.5 * p.tp * sigma_z

    return ham


def ham_weyl_2d_4bands(k, params=None):
    """Four-band hamiltonian of a Weyl semimetal without
    time-reversal symmetry.

    Taken from  `https://arXiv.org/abs/1303.5784` (equivalent to
    `https://arxiv.org/abs/1105.5138` up to unitary transformations).

    Parameters
    ----------
    params: dict
        Can contain the parameters
        layers: int, default to `1`
            Number of layers in the slab.
        pbc : bool, default to False
            If 'True', use periodic boundary conditions, and the slab will
            not host surface states.
        p: SimpleNamespace, optional
            Contains the Hamiltonian parameters,
                p.mu : chemical potential
                p.t : (spin-orbit) hopping parameter
                p.tp : kinetic hopping parameter
                p.m0 : constant mass term (m0<0 --> topological phase)
                p.k0 : half the k-space separation between Weyl nodes
                p.lam : energy shift between Weyl nodes
    """
    if params is None:
        params = {'layers': 1}
    elif len(params) == 0:
        raise ValueError('`params` must be provided.')
    layers = params['layers']
    pbc = params.get('pbc', False)
    assert layers > 0

    t0s0 = np.kron(tau_0, sigma_0)
    tZsX = np.kron(tau_z, sigma_x)
    tZsY = np.kron(tau_z, sigma_y)
    tZsZ = np.kron(tau_z, sigma_z)
    tXs0 = np.kron(tau_x, sigma_0)
    t0sZ = np.kron(tau_0, sigma_z)
    tZs0 = np.kron(tau_z, sigma_0)

    kx, ky = k
    coskx = np.cos(kx)
    cosky = np.cos(ky)
    sinkx = np.sin(kx)
    sinky = np.sin(ky)

    try:
        p = params['p']
    except KeyError:
        p = SimpleNamespace(mu=0.001, t=1, tp=1, m0=0., k0=1.2, lam=0.2)
    ham_dim = 4 * layers

    m = p.m0 + p.tp * (1 - coskx) + p.tp * (1 - cosky) + p.tp

    ham = np.zeros((ham_dim, ham_dim), dtype=complex)
    s = [slice(4*i, 4*i+4) for i in range(layers)]
    for i in range(layers):
        ham[s[i], s[i]] = (
            p.t * sinkx * tZsY + p.t * sinky * tZsZ
            + m * tXs0 + p.k0 * t0sZ - p.mu * t0s0 - p.lam * tZs0)
    for i in range(layers - 1):
        ham[s[i], s[i+1]] = -0.5j * p.t * tZsX - 0.5 * p.tp * tXs0
        ham[s[i+1], s[i]] = 0.5j * p.t * tZsX - 0.5 * p.tp * tXs0

    if pbc:
        ham[s[-1], s[0]] = -0.5j * p.t * tZsX - 0.5 * p.tp * tXs0
        ham[s[0], s[-1]] = 0.5j * p.t * tZsX - 0.5 * p.tp * tXs0

    return ham
