from matplotlib import pyplot as plt
import scipy.optimize
import numpy as np


def psi(ms, b=0):
    """Returns the wavefuctions for all k-points separated in spins.

    The output has shape '(num_k, num_layers, 2)'."""
    wfs = ms.wf
    num_k, num_layers = wfs[b].shape
    return wfs[b].reshape(num_k, num_layers // 2, 2)  # separate spins


def psi_sqr(ms, b=0):
    """Returns the probability densities for all k-points.

    The output has shape '(num_k, num_layers)'."""
    psi_array = psi(ms, b=b)
    return np.linalg.norm(psi_array, axis=-1) ** 2


def bulk_weight(ms, b=0, thickness=1):
    probs = psi_sqr(ms, b=b)
    return np.sum(probs[:, thickness:-thickness], axis=1)


def weight_lower(ms, b=0, thickness=1):
    probs = psi_sqr(ms, b=b)
    return np.sum(probs[:, :thickness], axis=1)


def weight_upper(ms, b=0, thickness=1):
    probs = psi_sqr(ms, b=b)
    return np.sum(probs[:, -thickness:], axis=1)


def surface_weight(ms, b=0, thickness=1):
    return (weight_upper(ms, b=b, thickness=thickness) +
            weight_lower(ms, b=b, thickness=thickness))


def get_colors_contour(ms, band_idx=0, thickness=4):
    """Color bands according to surface weight.

    'green' is majority of bulk weight, 'blue' for top surface,
    and 'red' for bottomg surface."""
    bands = ms.bands
    top_weight = weight_upper(ms, b=band_idx, thickness=thickness)
    bottom_weight = weight_lower(ms, b=band_idx, thickness=thickness)
    center_weight = bulk_weight(ms, b=band_idx, thickness=thickness)
    colors = np.ndarray(len(bands[band_idx]), dtype=object)
    colors[top_weight > bottom_weight] = 'blue'
    colors[top_weight < bottom_weight] = 'red'
    colors[center_weight > top_weight + bottom_weight] = 'green'
    return colors


def plot_abs(ms, bands, surf_idx, thickness=2):
    nlayers = ms.wf[surf_idx].shape[1] // 2
    z = np.arange(nlayers)
    zp = z - nlayers // 2

    fig, ax = plt.subplots(figsize=(3.5, 3))

    psisq = np.zeros_like(z)
    for bi in range(len(bands)):
        if bi != surf_idx:
            psisq = np.max([psisq, *psi_sqr(ms, bi)], axis=0)
        ax.fill_between(zp, psisq, color='lightgray',  edgecolor='gray')

    psisq = psi_sqr(ms, surf_idx)
    colors = get_colors_contour(ms, band_idx=surf_idx, thickness=thickness)
    for _psi, col in zip(psisq, colors):
        ax.plot(zp, _psi, c=col, lw=1, alpha=0.6)

    ax.tick_params(labelsize=10)

    ax.set_xlim(zp[0], zp[-1])

    ax.set_xlabel('$z\,[a]$')
    ax.set_ylabel('$|\psi|^2$', rotation=0, labelpad=16)

    return fig


def plot_bands(bands, col, ax=None, fs=12, s=None, every=1):
    if s is None:
        s = fs / 2
    fig = plt.gcf()
    if ax is None:
        ax = fig.subplots(1, 1)

    # plot bulk bands
    for band in bands[1:]:
        xy = band.transpose()
        ax.scatter(xy[0, ::every], xy[1, ::every], c='gray', lw=0, s=s)

    # plot surface bands
    xy = bands[0].transpose()
    ax.scatter(xy[0, ::every], xy[1, ::every], c=col[::every], lw=0, s=s)

    ax.set_xlabel('$k_x\,[\pi/a]$', fontsize=fs)
    ax.set_ylabel('$k_y\,[\pi/a]$', fontsize=fs)
    ax.tick_params(labelsize=fs)

    ax.set(aspect=1)


def power_law(x, p, A, B):
    return A*x**p + B


def linear_func(x, a, b):
    return a * x + b + 1


def inv_linear_func(x, a, b):
    return a / x + b


def plot_sigma_xx_yy(xi_array, cond_array, fitting=True, axs=None,
                     title_fs=9, fs=8, markersize=4,
                     fitting_slice=slice(11, None), p0=(1, 1, 0)):

    fig = plt.gcf()
    if axs is None:
        axs = fig.subplots(1, 2)

    ax1, ax2 = axs

    title_fs = 12
    fs = 10

    ax1.plot(xi_array[:], cond_array[:, 0, 0], 'o', markersize=markersize)

    ax1.set_title('$\sigma_{xx}/\sigma_{0}$', fontsize=title_fs)
    ax1.set_xlabel('$ \\xi\,[a] $', fontsize=fs)

    ax2.plot(xi_array[:], cond_array[:, 1, 1], 'o', markersize=markersize)
    ax2.set_title('$\sigma_{yy}/\sigma_{0}$', fontsize=title_fs)
    ax2.set_xlabel('$ \\xi\,[a] $', fontsize=fs)

    ax1.tick_params(labelsize=fs)
    ax2.tick_params(labelsize=fs)

    if fitting:
        fit, pcov = scipy.optimize.curve_fit(power_law,
                                             xi_array[fitting_slice],
                                             cond_array[fitting_slice, 0, 0],
                                             p0=p0)
        ax1.plot(xi_array[:], power_law(xi_array[:], *fit), c='green', lw=1.2)
        xi_0 = 1
        ax1.annotate(
            '$'+'\\propto\\xi^{'+str(np.round(fit[0], 1))+'}$',
            (xi_0, (power_law(xi_0, *fit) +
                    power_law(xi_array[-1], *fit)/5)),
            color='green')
        fit, pcov = scipy.optimize.curve_fit(power_law,
                                             xi_array[fitting_slice],
                                             cond_array[fitting_slice, 1, 1],
                                             p0=p0)
        ax2.plot(xi_array[:], power_law(xi_array[:], *fit), c='green', lw=1.2)
        ax2.annotate(
            '$'+'\\propto\\xi^{'+str(np.round(fit[0], 1))+'}$',
            (xi_0, (power_law(xi_0, *fit) +
                    power_law(xi_array[-1], *fit) / 5)),
            color='green')


def plot_cond_vs_w(w_array, cond_array, axs=None,
                   fs=10, title_fs=12, markersize=2,
                   fit='', annotate=True):
    fig = plt.gcf()
    if axs is None:
        ax1, ax2 = fig.subplots(1, 2)
    else:
        ax1, ax2 = axs

    cxx = cond_array[:, 0, 0]
    cyy = cond_array[:, 1, 1]

    ax1.plot(w_array, cxx, 'o', c='C0',  markersize=markersize)
    ax2.plot(w_array, cyy, 'o', c='C0', markersize=markersize)

    if fit == 'inv':
        popt1, pcov = scipy.optimize.curve_fit(
            inv_linear_func, w_array[:], cxx)
        ax1.plot(w_array, inv_linear_func(w_array, *popt1), '--', c='g', lw=1)
        if annotate:
            ax1.annotate(
                '$\propto W^{-1}$', (w_array[3] + 1e0,
                                     inv_linear_func(w_array[3], *popt1)),
                xycoords='data', rotation=0, color='g', annotation_clip=True)

        popt2, pcov = scipy.optimize.curve_fit(
            inv_linear_func, w_array[:], cyy)
        ax2.plot(w_array, inv_linear_func(w_array, *popt2), '--', c='g', lw=1)
        if annotate:
            ax2.annotate(
                '$\propto W^{-1}$', (w_array[3] + 1e0,
                                     inv_linear_func(w_array[3], *popt2)),
                xycoords='data', rotation=0, color='g',
                annotation_clip=True)
    elif fit == 'linear':
        popt1, pcov = scipy.optimize.curve_fit(
            linear_func, w_array[:], cxx)
        ax1.plot(w_array, linear_func(w_array, *popt1), '--', c='g', lw=1)
        if annotate:
            ax1.annotate(
                '$\propto W$', (w_array[3] + 1e0,
                                linear_func(w_array[3], *popt1)),
                xycoords='data', rotation=0, color='g',
                annotation_clip=True)

        popt2, pcov = scipy.optimize.curve_fit(
            linear_func, w_array[:], cyy)
        ax2.plot(w_array, linear_func(w_array, *popt2), '--', c='g', lw=1)
        if annotate:
            ax2.annotate(
                '$\propto W$', (w_array[3] + 1e0,
                                linear_func(w_array[3], *popt2)),
                xycoords='data', rotation=0, color='g',
                annotation_clip=True)

    ax1.set_title('$\sigma_{xx}/\sigma_{0}$', fontsize=title_fs)
    ax2.set_title('$\sigma_{yy}/\sigma_{0}$', fontsize=title_fs)

    ax1.set_xlabel('$ W\,[a] $', color='k', fontsize=fs)
    ax1.tick_params(labelsize=fs)

    ax2.set_xlabel('$ W\,[a] $', color='k', fontsize=fs)
    ax2.tick_params(labelsize=fs)


def plot_vp_vs_w(w_array, valley_polarization, fitting=True, ax=None,
                 title_fs=9, fs=8, markersize=4, fitting_slice=slice(11, None),
                 p0=(1, 1, 0), ylabel=True):

    fig = plt.gcf()
    if ax is None:
        ax = fig.add_subplot()

    title_fs = 12
    fs = 10

    ax.plot(w_array, valley_polarization, 'o', markersize=markersize)

    if ylabel:
        ax.set_ylabel('$\chi$', fontsize=fs)
    else:
        ax.set_title('$\chi$', fontsize=title_fs)
    ax.set_xlabel('$W\,[a]$', fontsize=fs)

    ax.tick_params(labelsize=fs)

    if fitting:
        fit, pcov = scipy.optimize.curve_fit(
            power_law, xi_array[fitting_slice],
            valley_polarization[fitting_slice], p0=p0)
        ax.plot(xi_array[:], power_law(xi_array[:], *fit), c='green', lw=1.2)
        xi_0 = 1
        ax.annotate('$'+'\\propto W^{'+str(np.round(fit[0], 1))+'}$',
                    (xi_0, (power_law(xi_0, *fit) +
                            power_law(xi_array[-1], *fit) / 5)),
                    color='green')


def plot_vp_vs_xi(xi_array, valley_polarization, fitting=True, ax=None,
                  title_fs=9, fs=8, markersize=4,
                  fitting_slice=slice(11, None), p0=(1, 1, 0), ylabel=True):

    fig = plt.gcf()
    if ax is None:
        ax = fig.add_subplot()

    title_fs = 12
    fs = 10

    ax.plot(xi_array, valley_polarization, 'o', markersize=markersize)

    if ylabel:
        ax.set_ylabel('$\chi$', fontsize=fs)
    else:
        ax.set_title('$\chi$', fontsize=title_fs)
    ax.set_xlabel('$\\xi\\,[a]$', fontsize=fs)

    ax.tick_params(labelsize=fs)

    if fitting:
        fit, pcov = scipy.optimize.curve_fit(
            power_law, xi_array[fitting_slice],
            valley_polarization[fitting_slice], p0=p0)
        ax.plot(xi_array[:], power_law(xi_array[:], *fit), c='green', lw=1.2)
        xi_0 = 1
        ax.annotate('$'+'\\propto \\xi^{'+str(np.round(fit[0], 1))+'}$',
                    (xi_0, (power_law(xi_0, *fit) +
                            power_law(xi_array[-1], *fit) / 5)),
                    color='green')
